

package dumb_line_re_lexer;

use strict;

my $MAX_LINE_SIZE_DEFAULT = 1024;


sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};
    bless ($self, $class);

    my %args = @_;

    $self->{pattern_table} = $args{pattern_table};

    $self->{line_buffer} = undef;
    $self->{MAX_LINE_SIZE} = $MAX_LINE_SIZE_DEFAULT;

    $self->{stream} = $args{stream}; 

    return $self;
}


sub get_token {

    # Writing this because I could not persuade Parse::Lex to not allow newlines in strings,
    # and because I am too lazy to write a Lex in Perl.
    # Probably acceptable for limited interactive parsing of SQL,
    # in my anticipated application;
    # unsuited for major SQL work.

    my($self) = @_;

    for(;;) {
	# (0) repeat the following until a token is found, or the input ends

	# (1) get a line worth of data from stream (by repeatedly calling getc)
	#     error if exceeds buffer length
	
	if( !defined $self->{line_buffer} || $self->{line_buffer} eq "" ) {
	    $self->{line_buffer} = getline($self);
	    if( !defined $self->{line_buffer} ) {return undef;}
	}
	
	# (2) naively apply patterns to the beginning of the string.
	#     patterns are not allowed to cross a line boundary

	my $match_pt = undef;
	my $match_len = 0;
	my $match_string = undef;
	my $match_remainder = undef;
	my $pattern_tuple;
	foreach $pattern_tuple ( @{$self->{pattern_table}} ) {
	    if( $self->{line_buffer} =~ m/^($pattern_tuple->{regexp})(.*)/sx ) {
		my $this_match_len = length($1);
		if( !defined $match_pt || $this_match_len > $match_len ) {
		    $match_len = $this_match_len;
		    $match_pt = $pattern_tuple;
		    $match_string = $1;
		    $match_remainder = $2;
		}
	    }
	}
	if( $match_pt && ( $match_len > 0 || $match_remainder eq '' ) ) {
	    if( $match_pt->{action} eq 'IGNORE' ) {
		# do nothing, except move on (below)
		$self->{line_buffer} = $match_remainder;
	    }
	    elsif( $match_pt->{action} eq 'RETURN' ) {
		$self->{line_buffer} = $match_remainder;
		my %match = %$match_pt;
		$match{string} = $match_string;
		return \%match;
	    }
	    elsif( $match_pt->{action} =~ '^ERROR' ) {
		die "$match_pt->{action}: $self->{line_buffer}\n";
	    }
	    else {
		die "Unknown or missing action: " . $match_pt->{action}||"undefined"  . ": $self->{line_buffer}";
	    }
	}
	else {
	    die "No lexical pattern match: <$self->{line_buffer}>";
	}
	
    }
    
}

sub getline {
    my($self) = @_;

    my $buf = "";
    my $len = 0;

    for(;;) {
	my $c = $self->{stream}->getc;
	if( !defined $c ) {
	    if( $buf eq "" ) {return undef;}
	    else {return $buf;}
	}
	$buf .= $c;
	if( $c eq "\n" ) {return $buf;}
	if( ++$len > $self->{MAX_LINE_SIZE} ) {
	    die "Line too long: > $self->{MAX_LINE_SIZE} characters: $buf\n";
	}
    }
}



