#!/s/perl-5.004/bin/perl -w

## $Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/plain_text_database.pm,v 1.3 1998/05/12 19:35:51 glew Exp $

##
## plain_text_database
## 
## tools to manipulate "databases" 
## in plain text files, of several simple formats.
## 
## Basic object is a tuple that points to a nice_IO handle 
## as well as things like file (table) name,
## format, flags to indicate whether and how sorted,
## etc.
## 
## Unclear whether at this point I want the database file (table)
## object to have a lifetime longer than that of a single open
## instance - do I want to cache information across closes?
## 
## Methods include:
##
## (0.TBD) Opening and closing a databse file (table)
## (1.TBD) Figuring out format of a database file by inspection
## (2.TBD) input scans that getrecord one at a time
## (3.TBD) output scans that putrecord one at a time
## (4.TBD) getrecords/putrecords to read all into memory from file
## (5.TBD) Database algorithms, including:
## (5.1.TBD) SELECTs (simple scans, etc.)
## (5.2.TBD) UPDATES
## (5.3.TBD) INNER JOINs (on sorted files)
## (5.4.TBD) OUTER JOINs (on sorted files)
## etc.
##
## TBD: It is hoped one day to integrate this with a Perl SQL parser.
## The naive parser from Sprite may be a start.
##
## TBD: It is hoped one day to integrate this with a query optimizer.
## 
## TBD: extend to accept Perl internal hashes and arrays of hashes 
## as database tables.
## 

use strict;

package plain_text_database;

# Deduce location of libraries distributed with package
use FindBin;
BEGIN
{
    my $install_root = $FindBin::RealBin;
    $install_root =~ s:(.*/perl-sql)/.*:$1:;
    unshift (@INC, "$install_root/lib");
}

use Carp;

use Perl_SQL::nice_IO;
use Perl_SQL::null_handling;

use Perl_SQL::multiline_named_tuples;
use Perl_SQL::singleline_named_tuples;
use Perl_SQL::singleline_delimited_records;

use Perl_SQL::nice_IO;

use English;

use Data::Dumper;
use strict;


#
# CONSTRUCTOR
#
# Create a plain_text_database object
# linking it to an open file.
# Essentially, this is a cursor object,
# since it scans.
#
sub new {			# TBD: I would prefer to call it open
    my $proto = shift @_;
    my $class = ref($proto) || $proto;
    my $self  = {};
    bless ($self, $class);

    if( !defined @_ ) {
	warn "usage: open plain_text_database file-table-specifier...\n";
	$self->print_format_info_header($main::stdout);
	confess;
    }

    my $database_name = shift @_;

    $self->{NICE_IO_FILEHANDLE} = new nice_IO ($database_name,@_);
    if( !defined($self->{NICE_IO_FILEHANDLE}) ) {
	return undef;
    }

    $self->{DATABASE_OPEN_OPTIONS} = [$database_name,@_];

    # default settings
    # TBD: make the defaults class variables?
    # TBD: make the defaults user settable at constructor invocation time?
    $self->{FORMAT} = "unknown-database-format";
    $self->field_delimiter("whitespace");
    $self->make_consistent_database();
    
    return $self;
}

#
# explicitly specify format of an existing database
# TBD: overriding things already specified, or error?
#
sub database_format {
    my $self = shift;
    my %args = @_;

    my %key_copying_rules = (
			     # -1 => don't copy
			     # TBD: add other copying rules,
			     # such as always copy, copy if not already there, warn, etc.
			     NICE_IO_FILEHANDLE => -1,
			     DATABASE_OPEN_OPTIONS => -1,
			     );

    # $self is a partially built database handle.
    # for now, blindly copy everything in argument over.
    # TBD: may need to delete some stuff in $self;
    # for now, just create a new database handle if necessary.

    # usage:   
    #     $dbh = new plain_text_database ">+filename"; 
    #     $dbh->database_format( FORMAT => multiline_named_tuples, etc.)

    my $k;
    foreach $k ( keys %args ) {
	if( exists $key_copying_rules{$k} && $key_copying_rules{$k} == -1 ) {
	    # don't copy this field
	}
	else {
	    $self->{$k} = $args{$k};
	}
    }
    # default settings
    $self->make_consistent_database();
    return;
}

# attribute access methods, with coordinated actions
sub comment {
    my $self = shift;
    my $old = $self->{COMMENT};
    if (@_) { 
	$self->{COMMENT} = shift;
	$self->{NICE_IO_FILEHANDLE}->comment($self->{COMMENT});
    }
    return $old;
}
sub field_delimiter {
    my $self = shift;
    my $old = $self->{FIELD_DELIMITER};
    if (@_) { 
	$self->{FIELD_DELIMITER} = shift;
	if( $self->{FIELD_DELIMITER} eq "whitespace" ) {
	    $self->{FIELD_DELIMITER} = "\t";
	    $self->{IGNORE_WHITESPACE} = 1;
	}
	elsif( $self->{FIELD_DELIMITER} eq "tab" ) {
	    $self->{FIELD_DELIMITER} = "\t";
	    $self->{IGNORE_WHITESPACE} = 0;
	}

	$self->make_consistent_field_delimiter_regexp();
    }
    return $old;
}
sub ignore_whitespace {
    my $self = shift;
    my $old = $self->{IGNORE_WHITESPACE};
    if (@_) { 
	$self->{IGNORE_WHITESPACE} = shift;
	$self->make_consistent_field_delimiter_regexp();
    }
}
sub make_consistent_field_delimiter_regexp {
    my $self = shift;
    # TBD: allow user to explicitly specify FIELD_DELIMITER_REGEXP
    # separate from FIELD_DELIMITER?
    if( defined $self->{FIELD_DELIMITER} ) {
	if( $self->{IGNORE_WHITESPACE} ) {
	    if( $self->{FIELD_DELIMITER} eq ' '
	       || $self->{FIELD_DELIMITER} eq "\t"
	       ) {
		$self->{FIELD_DELIMITER_REGEXP} = '\s+';
	    }
	    else {
		$self->{FIELD_DELIMITER_REGEXP} = '\s*' . $self->{FIELD_DELIMITER} . '\s*';
	    }
	}
	else {
	    $self->{FIELD_DELIMITER_REGEXP} =  $self->{FIELD_DELIMITER};	
	}
    }
    else {
	# do not set FIELD_DELIMITER_REGEXP, since no FIELD_DELIMITER
	# TBD: set to whitespace?
    }
}

# TBD: other attributes - removing make_consistent_database?
# TBD: field_delimiter + whitespace -> regexp
# TBD: whitespace -> nice_IO whitespace
# TBD: type -> read_record and print_record method selection

sub make_consistent_database {
    my($self) = shift @_;

    $self->make_consistent_field_delimiter_regexp();

    $self->{NICE_IO_FILEHANDLE}->chomp(1); # TBD: always remove newline at end of lines?

    # make guesses/settings consistent

    # (0) set default I/O methods
    if( !defined $self->{read_record} ) {
	$self->{read_record} = \&plain_text_database::unknown_database_format_error;
    }
    if( !defined $self->{print_record} ) {
	$self->{print_record} = \&plain_text_database::unknown_database_format_error;
    }

    # (0) override according to format specified
    if( !defined $self->{FORMAT} ) {
	# TBD: Not known here, but not an error - user may provide
    }
    elsif( $self->{FORMAT} eq "multiline_named_tuples" ) {
	$self->{read_record} = \&multiline_named_tuples::read_record;
	$self->{print_record} = \&multiline_named_tuples::print_record;
	$self->{IGNORE_WHITESPACE} = 0;
    }
    elsif( $self->{FORMAT} eq "singleline_named_tuples" ) {
	$self->{read_record} = \&singleline_named_tuples::read_record;
	$self->{print_record} = \&singleline_named_tuples::print_record;
	$self->{IGNORE_WHITESPACE} = 1;
    }
    elsif( $self->{FORMAT} eq "singleline_delimited_records" ) {
	$self->{read_record} = \&singleline_delimited_records::read_record;
	$self->{print_record} = \&singleline_delimited_records::print_record;
	# Don't change $self->{IGNORE_WHITESPACE}, since format dependent
    }
    else {
	# TBD: Not known here, but not an error - user may provide
    }

    if( $self->{IGNORE_WHITESPACE} ) {
	$self->{NICE_IO_FILEHANDLE}->remove_whitespace(1);
    }
    else {
	$self->{NICE_IO_FILEHANDLE}->remove_whitespace(0);
    }

}

sub unknown_database_format_error {
    my $self = shift;
    
    print STDERR "error: unknown database type: " . ($self->{FORMAT} || 'undef') . "\n";
    print STDERR "error: probably caused by poorly formatted database file (or bad code)\n";
    exit(1);			# using standard "die" causes test diffing problems
}

#
# Try to guess what file format is used for a plain text database
# TBD: guess, but return guess instead of setting it?
#
sub guess_format {
    my $self = shift;
    
    # the real work is done by the helper routine
    $self->guess_format_helper();

    $self->make_consistent_database();

    return $self->{FORMAT};
}

sub guess_format_helper {
    my $self = shift;
    # TBD: use rest of command line as default?
    # TBD: use filename *.tsv *.csv as default?
    # TBD: use global (or class) variables as default?

    my $nfh = $self->{NICE_IO_FILEHANDLE};

    my $old_filter = $nfh->filter(0);

    # Heuristic for figuring out comment character
    # ++ find first non-blank line
    # ++ see if it is one of the common comment characters
    for(;;) {
	my $line;
	if( !defined($line = $nfh->getline()) ) { last; }
	chomp $line;
	# find non-blank line
	if( $line !~ m/^\s*$/ ) { 
	    # test against common comments
	    if( $line =~ m/^\s*#/ ) {
	       $self->comment('#');
	    }
	    elsif( $line =~ m:^\s*//: ) {
	       $self->comment('//');
	    }
	    elsif( $line =~ m:^\s*!: ) {
	       $self->comment('!');
	    }
	    $nfh->ungetlines($line);
	    last;
	}
    }
    my $comment = $self->comment();
    


    # Heuristic for detecting delimiters and fieldnames
    # ++ assume # is comment (TBD: fix)
    # ++ read (saving) until we come to first 2 non-blank non-comment lines
    # ++ inspect these

    my @saved_header_lines;
    my $significant_line_1;
    my $significant_line_2;

    my $line;

    for(;;) {
	if( !defined($line = $nfh->getline()) ) { last; }
	chomp $line;
	@saved_header_lines = ( @saved_header_lines, $line );
	if( $line =~ m/^\s*$/ ) {
	    # blank. save.
	}
	elsif( defined $comment && $line =~ m/^\s*$comment.*/ ) {
	    # comment. save and process.
	    if( $line =~ m/^\s*$comment\s*field-names:\s*(.*)/i 
	       or  $line =~ m/^\s*$comment\s*fieldnames:\s*(.*)/i 
	       or  $line =~ m/^\s*$comment\s*fields:\s*(.*)/i 
	       ) {
	       my @fieldnames = split(/,\s*/,$LAST_PAREN_MATCH);
	       $self->{FIELD_NAMES} = \@fieldnames;
	    }
	    elsif( $line =~ m/^\s*$comment\s*field-delimiter:\s*(.*)/ ) {
		$self->field_delimiter($LAST_PAREN_MATCH);
	    }
	    elsif( $line =~ m/^\s*$comment\s*comment:\s*(.*)/ ) {
		$self->comment($LAST_PAREN_MATCH);
	    }
	    elsif( $line =~ m/^\s*$comment\s*format:\s*(.*)/ ) {
		$self->{FORMAT} = $LAST_PAREN_MATCH;
		  # TBD: validate format against supported list...
	    }
	    elsif( $line =~ m/^\s*$comment\s*sorted-by:\s*(.*)/ ) {
		$self->{SORTED} = $LAST_PAREN_MATCH;
	    }
	    elsif( $line =~ m/^\s*$comment\s*ignore-whitespace:\s*(.*)/ ) {
		$self->{IGNORE_WHITESPACE} = $LAST_PAREN_MATCH;
	    }
	}
	elsif( !defined $significant_line_1 ) {
	    # non-blank, non-comment.
	    $significant_line_1 = $line;
	} else {
	    $significant_line_2 = $line;
	    last;
	}
    }
    $nfh->ungetlines(@saved_header_lines);
    if( defined $self->{FORMAT} && ($self->{FORMAT} eq "multiline_named_tuples") ) {
	# this format is self-schematizing - we are done
	goto return_restoring_filter;
	# TBD: erase or warn about other definitions?
    }
    if( defined $self->{FORMAT} && ($self->{FORMAT} eq "singleline_named_tuples") ) {
	# this format is self-schematizing - we are done
	goto return_restoring_filter;
	# TBD: erase or warn about other definitions?
    }
    if( defined $self->{FORMAT} && ($self->{FORMAT} eq "singleline_delimited_records") ) {
	# this format is not self-schematizing,
	# but we will assume that the user has set things in the comments,
	# or that we will use defaults.
	# TBD: use filename to set defaults?
	goto return_restoring_filter;
	# TBD: erase or warn about other definitions?
    }
    if( defined $self->{FIELD_NAMES} ) {
	# guess that we are SDR
	$self->{FORMAT} = "singleline_delimited_records";
	# don't return yet: may need to figure out parameters
	# and may override assumption of SDR
	# TBD: skip the heuristic if the parameters are already defined
    }
    if( !defined $significant_line_1 ) {
	goto return_restoring_filter;
    }
    if( $significant_line_1 =~ m/\s*\{\s*$/ ) {
	$self->{FORMAT} = "multiline_named_tuples";
	goto return_restoring_filter;
    }
    if( $significant_line_1 =~ m/\s*^\{.*\}(,|)\s*$/ ) {
        $self->{FORMAT} = "singleline_named_tuples";
	goto return_restoring_filter;
    }
    if( !defined $significant_line_2 ) {
	guess_check_likely_delimiters($self,$significant_line_1);
    }
    guess_check_likely_delimiters($self,$significant_line_1);
    my($sep1,$ws1) = ($self->{FIELD_DELIMITER},$self->{IGNORE_WHITESPACE});   
    guess_check_likely_delimiters($self,$significant_line_1);
    my($sep2,$ws2) = ($self->{FIELD_DELIMITER},$self->{IGNORE_WHITESPACE});   
    if( defined $sep1 ) {
	if( $sep1 ne $sep2 ) {
	    goto return_restoring_filter;
	}
	$self->{IGNORE_WHITESPACE} = $ws1 | $ws2;

	if( ! defined $self->{FIELD_NAMES} ) {
	    # now check to see if first significant (non-comment)line consists of field names
	    my $fn;
	    $self->make_consistent_database(); # create regexps, etc., before more parsing of fields.
	    my @field_names = split(/$self->{FIELD_DELIMITER_REGEXP}/,$significant_line_1);
	    foreach $fn ( @field_names ) {
		$fn =~ s/\s+//g;
		if( $fn !~ m/^[a-zA-Z_]/ ) {
		    goto return_restoring_filter;
		}
	    }
	    # it *is* fieldnames: don't push back header line(s)
	    $self->{FIELD_NAMES} = \@field_names;
	    $nfh->ungetlines($significant_line_2);
	    $self->{FORMAT} = "singleline_delimited_records";
	    goto return_restoring_filter;
	}
	
    }
    goto return_restoring_filter;

return_restoring_filter:

    $self->make_consistent_database();

    $nfh->filter($old_filter);

    if( $self->{FORMAT} eq "unknown-database-format" ) {
	return undef;
    }
    else {
	return $self->{FORMAT};
    }

}





sub guess_check_likely_delimiters {
    my $self = shift @_;
    my($line) = @_;
    my $c;
    
    # Scan over likely explicit delimiters,
    # both with and without whitespace elision
    foreach $c ( split(/ */, ',;:') ) {
	my $result = guess_check_likely_delimiter_character($self,$c,$line);
	if( defined $result ) { 
	    return $result; 
	}
    }
    # Variable Whitespace Delimiting (VWD)
    # single blank, or two adjacent tabs
    if( $line =~ m/\S+( |\t\t)\s*\S+/ ) {
	# variably whitespace separated
	# TBD: only a single blank is necessary to get into this space. 
	# No non variable blank separated records?
	$self->field_delimiter("whitespace");
	return $self->{FORMAT} = "singleline_delimited_records";
    }
    # Tab Delimited (TSV)
    if( $line =~ m/\S+\t\S+/ ) {
	# strictly tab separated
	$self->field_delimiter("tab");
	$self->{IGNORE_WHITESPACE} = 0;
	return $self->{FORMAT} = "singleline_delimited_records";
    }
    # none found
    return undef
}
	    

sub guess_check_likely_delimiter_character {
    my $self = shift @_;
    my ($char,$line) = @_;

    if( $line =~ m/\S+\s*$char\s+\S+/ 
       || $line =~ m/\S+\s+$char\s*\S+/ 
       ) {
	# if has at least one whitespace around delimiter, anywhere,
	# treat it as whitespace ignored delimited
	$self->field_delimiter($char);
	$self->ignore_whitespace(1);
	return $self->{FORMAT} = "singleline_delimited_records";
    }
    if( $line =~ m/\S+$char\S+/ ) {
	# has delimiters, but no whitespace
	$self->field_delimiter($char);
	$self->ignore_whitespace(0);
	return $self->{FORMAT} = "singleline_delimited_records";
    }
    return undef;
}

sub get_next_tuple {
    my $self = shift @_;
    my $tupleref = &{$self->{read_record}}($self);
    #print "debug returning from get_next_tuple: ", Dumper(\%tup);
    return $tupleref;
}


sub put_next_tuple {
    my $self = shift;
    my $tuple = shift;
    return &{$self->{print_record}}($self,$tuple);
}

# print raw information to database 
# such as format header
#  usage:  $ptd->print_raw( $ptd->format_to_string )
# can also be used for other comments
sub print_raw {
    my $self = shift;
    $self->{NICE_IO_FILEHANDLE}->print( @_ );
}


sub print_format {
    my $self = shift;
    die "unexpected arguments in plain_text_database::print_format\n" if @_;
    $self->print_raw( $self->format_to_string() );
}

# print the formatting info
# out as the comment header for a file.
# also useful for debugging
sub format_to_string {
    my $self = shift @_;
    my %options = @_;
    my $result = "";

    if( defined $options{date} && $options{date} eq 0 ) {
	# skip printing date, mainly to get easily reproducible test output
    }
    else {
	my $date = `date`;
	chomp $date;
	$result .= "# date:                    " . $date . "\n";
    }

    my $comment;
    if( defined $self->{'COMMENT'} ) {
	$comment = $self->{'COMMENT'};
    }
    else {
	# TBD: unclear if should use comment, or nothing.
	$comment = '#';
    }

    # TBD: use a naming convention so that I can add more stuff easily
    # TBD: use a code for undefined values or fields

    if( defined($self->{DATABASE_OPEN_OPTIONS}) ) { 
	$result .= "${comment} database file opened as: "
	    . join(',',@{$self->{DATABASE_OPEN_OPTIONS}}) . "\n";
    }
    if( defined($self->{FORMAT}) ) { 
	$result .= "${comment} format:                  $self->{FORMAT}\n";
    }
    else {
	$result .= "${comment} unknown format\n";
    }
    if( defined($self->{FIELD_DELIMITER}) ) { 
	my $delim;
	if( $self->{FIELD_DELIMITER} eq "\t" ) {
	    if( $self->{IGNORE_WHITESPACE} ) {
		$delim = 'whitespace';
	    }
	    else {
		$delim = 'tab';
	    }
	}
	else {
	    $delim = $self->{FIELD_DELIMITER};
	}
	$result .= "${comment} field-delimiter:         $delim\n";
    }
    if( defined($self->{FIELD_NAMES}) ) { 
	$result .= "${comment} field-names:             " . join(', ',@{$self->{FIELD_NAMES}}) . "\n";
	# TBD: use a delimiter in field-name list?
    }
    if( $self->comment() ) { 
	# TBD: am not really using this - should use for header...
	$result .= "${comment} comment:                 " . $self->comment() . "\n";
    }
    if( defined($self->{IGNORE_WHITESPACE}) ) { 
	$result .= "${comment} ignore-whitespace:       $self->{IGNORE_WHITESPACE}\n";
    }
    if( defined($self->{SORTED}) ) { 
	$result .= "${comment} database file opened as: $self->{SORTED}\n";
    }
    return $result;
}


sub seek {
    my $self = shift;
    my $offset = shift;
    my $whence = shift;
    die "Too many arguments to plain_text_database::seek\n" if @_;
    
    return $self->{NICE_IO_FILEHANDLE}->seek($offset,$whence);
}

sub string_to_database_format {
    my $string = shift @_;

    if( 0 ) { 
	# skip
    }
    elsif( $string eq "multiline_named_tuples" 
	or $string =~ m/MNT/i
	) {
	    return { FORMAT => "multiline_named_tuples" };
	}
    elsif( $string eq "singleline_named_tuples" 
	or $string =~ m/SNT/i
	) {
	    return { FORMAT => "singleline_named_tuples" };
	}
    elsif( $string eq "singleline_delimited_records" 
	or $string =~ m/SDR/i
	) {
	    return 
	    { 
		FORMAT => "singleline_delimited_records",
		IGNORE_WHITESPACE => 1,
		FIELD_DELIMITER => ',',
	    };
	}
    elsif( $string eq "comma_separated_variables" 
	or $string =~ m/CSV/i
	) {
	    return 
	    { 
		FORMAT => "singleline_delimited_records",
		IGNORE_WHITESPACE => 1,
		FIELD_DELIMITER => ',',
	    };
	}
    elsif( $string eq "tab_separated_variables" 
	or $string =~ m/TSV/i
	) {
	    return 
	    { 
		FORMAT => "singleline_delimited_records",
		IGNORE_WHITESPACE => 0,
		FIELD_DELIMITER => "\t",
	    };
	}
    elsif( $string eq "whitespace_separated_variables" 
	or $string =~ m/WSV/i
	) {
	    return 
	    { 
		FORMAT => "singleline_delimited_records",
		IGNORE_WHITESPACE => 1,
		FIELD_DELIMITER => ' ',
	    };
	}
    else {
        die "unknown format $string\n";
    }
    

}

"KLUGE: Perl requires modules to return a true value";
